/** \file threader.h
 *
 * thread-specific definitions and prototypes.
 *
 */
#ifndef THREADER_H_
#define THREADER_H_

#include "tinetype.h"
#if defined(__cplusplus) && !defined(FORCE_CPP)
extern "C" {
#endif

#ifdef MULTITHREADED

typedef struct 
{
  void (*tsk)(void);
  int rate;
  int hasReference;
  void *elRef;
  void *EqmRef;
} BkgThreadFcn;

#if defined(UNIX)

#define STD_THREAD_PRIORITY 0
#define MIN_THREAD_PRIORITY -99
#define MAX_THREAD_PRIORITY 99
#define THREAD_TYPE void *
#define THREAD_PARAM void *
#include "pthread.h"
#include "semaphore.h"
typedef pthread_mutex_t * MXHANDLE;
typedef sem_t * SEMHANDLE;
typedef pthread_t THRHANDLE;
extern pthread_attr_t gBkgThreadAttr;
void *bkgThreadTask(void *lpvThreadParam);
#define threadExit(t) pthread_cancel((pthread_t)t)
#define ReleaseMutex(mx) pthread_mutex_unlock(mx)
#define ReleaseSemaphore(mx) sem_post(mx)

#elif defined(WIN32)

#define STD_THREAD_PRIORITY THREAD_PRIORITY_NORMAL
#define MIN_THREAD_PRIORITY THREAD_PRIORITY_LOWEST
#define MAX_THREAD_PRIORITY THREAD_PRIORITY_HIGHEST
#define THREAD_TYPE DWORD WINAPI 
#define THREAD_PARAM LPVOID
typedef HANDLE MXHANDLE;
typedef HANDLE SEMHANDLE;
typedef HANDLE THRHANDLE;
extern DWORD dwBkgThreadId;
DWORD WINAPI bkgThreadTask(LPVOID lpvThreadParam);
#define threadExit(t) TerminateThread((HANDLE)t,0)
#define bkgThreadResume(t) ResumeThread((HANDLE)t)

#elif defined(VXWORKS)
/* these priorities are relative to the task spawn priority */
#define STD_THREAD_PRIORITY 0
#define MIN_THREAD_PRIORITY -99
#define MAX_THREAD_PRIORITY 99
#define THREAD_TYPE int
#define THREAD_PARAM void *
typedef SEM_ID MXHANDLE;
typedef SEM_ID SEMHANDLE;
typedef int THRHANDLE;
THREAD_TYPE bkgThreadTask(THREAD_PARAM lpvThreadParam);
int vxwReleaseSemaphore(MXHANDLE mx);
#define ReleaseMutex(mx) vxwReleaseSemaphore(mx)
#define ReleaseSemaphore(mx) vxwReleaseSemaphore(mx)
#define threadExit(t) taskDelete(t)

#endif /* UNIX or WIN32 or VXWORKS */

typedef struct
{
  SOCKET sck;
  int type;
  SEMHANDLE sem;
} SckThreadInfo;

THREAD_TYPE sendStreamThreadTask(THREAD_PARAM lpvThreadParam);
THREAD_TYPE recvStreamThreadTask(THREAD_PARAM lpvThreadParam);
THREAD_TYPE stkThreadTask(THREAD_PARAM lpvThreadParam);
THREAD_TYPE eqmThreadTask(THREAD_PARAM lpvThreadParam);
THRHANDLE eqmCreateThread(THREAD_PARAM lpvThreadParam);
THREAD_TYPE clnThreadTask(THREAD_PARAM lpvThreadParam);
THREAD_TYPE srvThreadTask(THREAD_PARAM lpvThreadParam);
THREAD_TYPE schThreadTask(THREAD_PARAM lpvThreadParam);
THREAD_TYPE sckSendThreadTask(THREAD_PARAM lpvThreadParam);
THREAD_TYPE sckRecvThreadTask(THREAD_PARAM lpvThreadParam);
THREAD_TYPE glbRecvThreadTask(THREAD_PARAM lpvThreadParam);
THREAD_TYPE mcstRecvThreadTask(THREAD_PARAM lpvThreadParam);
THREAD_TYPE scvThreadTask(THREAD_PARAM lpvThreadParam);
int eqmInitMutexSet(void *lpvMutexParam);
int stkInitMutexSet(void);

extern SEMHANDLE hSendStreamSemaphore;
extern SEMHANDLE hSchedulerSemaphore;
extern SEMHANDLE hNotificationSemaphore;
extern MXHANDLE hSystemInitMutex;
extern MXHANDLE hSystemServerMutex;
extern MXHANDLE hSystemKernelMutex;
extern MXHANDLE hCycleTriggerMutex;
extern MXHANDLE hSystemClientMutex;
extern MXHANDLE hSystemGlobalsUpdateMutex;
extern MXHANDLE hSystemMCastUpdateMutex;
extern MXHANDLE hLinkTblMutex;
extern MXHANDLE hDpdLinkMutex;
extern MXHANDLE hLwdLinkMutex;
extern MXHANDLE hLinkLstMutex;
extern MXHANDLE hTcpBucketMutex;
extern MXHANDLE hLinkQueueMutex;
extern MXHANDLE hClientTableMutex;
extern MXHANDLE hLogMutex;
extern MXHANDLE hCsvMutex;
extern MXHANDLE hMfMutex;
extern MXHANDLE hHistoryTableMutex;
extern MXHANDLE hHistoryCycleMutex;
extern MXHANDLE hAlmTblMutex;

extern int gClnThreadRunning;
extern int gSrvThreadRunning;

extern int srvThreadPriority;
extern int clnThreadPriority;
extern int bkgThreadPriority;

void YieldThread(void);
int InitMutex(MXHANDLE *mx);
int InitSystemMutexSet(void);
int CloseSystemMutexSet(void);
int CloseClientMutexSet(void);
int WaitForMutex(MXHANDLE mx,int to);
int ReleaseSystemMutex(MXHANDLE mx);
int CloseSystemMutex(MXHANDLE mx);
int WaitForSemaphore(SEMHANDLE mx,int to);
int ReleaseSystemSemaphore(SEMHANDLE mx);
int FreeSystemSemaphore(SEMHANDLE mx);
int WaitForThreadTermination(THRHANDLE hThr);

THRHANDLE mcstRecvCreateThread(THREAD_PARAM lpvThreadParam);
THRHANDLE glbRecvCreateThread(THREAD_PARAM lpvThreadParam);
THRHANDLE sckRecvCreateThread(SOCKET sck,int type);
THRHANDLE sckSendCreateThread(SOCKET sck,int type,SEMHANDLE *semHandle);
THRHANDLE schCreateThread(THREAD_PARAM lpvThreadParam);
THRHANDLE stkCreateThread(THREAD_PARAM lpvThreadParam);
THRHANDLE recvStreamCreateThread(SOCKET sck,int type);
THRHANDLE bkgCreateThread(BkgThreadFcn *tsk);
THRHANDLE clnCycleCreateThread(THREAD_PARAM lpvThreadParam);
THRHANDLE srvCycleCreateThread(THREAD_PARAM lpvThreadParam);
THRHANDLE scavangerCreateThread(THREAD_PARAM lpvThreadParam);

#define WaitForSystemMutex(mx,to) WaitForMutex(mx,to)

#define MTX_PRIMARY 0
#define MTX_SECONDARY 1
#define MTX_CARRIER 2

#else /* not MULTITHREADED */

typedef int MXHANDLE;
typedef int SEMHANDLE;
typedef int THRHANDLE;
extern int hSystemInitMutex;
extern int hSystemServerMutex;
extern int hSystemKernelMutex;
extern int hCycleTriggerMutex;
extern int hSystemClientMutex;
extern int hSystemGlobalsUpdateMutex;
extern int hSystemMCastUpdateMutex;
extern int hLinkTblMutex;
extern int hDpdLinkMutex;
extern int hLwdLinkMutex;
extern int hLinkLstMutex;
extern int hTcpBucketMutex;
extern int hLinkQueueMutex;
extern int hClientTableMutex;
extern int hLogMutex;
extern int hCsvMutex;
extern int hMfMutex;
extern int hHistoryTableMutex;
extern int hHistoryCycleMutex;
extern int hAlmTblMutex;

#endif /* MULTITHREADED */

extern TINE_EXPORT int useMultiThreadedBackgroundTasks;
extern TINE_EXPORT int useMultiThreadedEquipmentFunctions;
extern TINE_EXPORT int useMultiThreadedStockFunctions;
extern THRHANDLE hStkThread;
extern THRHANDLE hGlbThread;
extern THRHANDLE hMcstThread;
#define thread_dbglog(txt) {\
    dbglog("[T1]: %s %s %s -> %s",\
    cl->contract.EqmName,cl->contract.EqmDeviceName,\
    cl->contract.EqmProperty,(txt));\
  }
#define MUTEX_WAIT_AUGMENT 3

extern int gSystemMutexSetInitialized;
int dumpThreadPriorities(void);

TINE_EXPORT void SystemSetCycleTimer(void);
/**
 * \fn TINE_EXPORT void SystemStartCycleTimer(void)
 * \brief Starts a TINE engine 'cycler' running (SystemCycle()) in a separate thread.
 * 
 * Available in multi-threaded shared-library builds, this routine starts a separate thread
 * where SystemCycle() is allowed to run in a tight loop.  Only one instance of the cycler thread 
 * is allowed to run in a single process.  On windows systems, the thread will also see to it
 * that windows messages are allowed to be processed.
 *
 * \note Using the Windows DLLs (tine32.dll or tine64.dll) this routine is automatically 
 * called when the DLL is loaded.  When the shared library libtinemt.so is linked against,
 * the routine is likewise automatically called when the library is loaded, but ONLY if
 * either a) the library is built with -DTINE_START_CYCLER=TRUE or if the environment 
 * variable TINE_START_CYCLER = TRUE is set.  Builds using the static libraries, 
 * tinemt.lib (WINDOWS) or libtinemt.a (UNIX) will 
 * need to include tcycler.c in the project in order to make use of this routine.
 *
 * \note The developer is always at liberty to manage SystemCycle() in any manner he sees fit.
 * and to, for instance, include other tasks which need to be cycled over together in the same
 * managed loop.  However, it should be guaranteed that SystemCycle() runs in a single thread
 * per process.
 *
 * \see SystemKillCycleTimer().
 */
TINE_EXPORT void SystemStartCycleTimer(void);
/**
 * \fn TINE_EXPORT void SystemKillCycleTimer(void)
 * \brief Stops the TINE engine 'cycler' thread.
 * 
 * Available in multi-threaded shared-library builds, this routine stops the separate thread
 * where SystemCycle() runs in a tight loop (started via a call to SystemSetCycleTimer(). 
 *
 * \see SystemStartCycleTimer().
 */
TINE_EXPORT void SystemKillCycleTimer(void);

/**
 * \fn TINE_EXPORT void SystemWaitCycleTimer(void)
 * \brief Waits for the TINE engine 'cycler' thread to stop running.
 * 
 * Available in multi-threaded shared-library builds, this routine waits for the cycler thread
 * to stop running before continuing.
 *
 * \see SystemStartCycleTimer().
 */
TINE_EXPORT void SystemWaitCycleTimer(void);

#if defined(__cplusplus) && !defined(FORCE_CPP)
}
#endif

#endif /* THREADER_H_ */
